--
-- Test Bench for IIR filter
--
library ieee;
use std.textio.all;
use ieee.numeric_std.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_textio.all;


entity iir_RTL_test_bench is
end entity iir_RTL_test_bench;


architecture test of iir_RTL_test_bench is
   -- Path
   constant path : string := "C:\User\Daniel\Protel\FPGA_Biquad\";
   -- Input file. Format: x
   constant sim_in : string (1 to 50) :=
         "simdata\stim_sqrsweep1_17s.txt####################";
   -- Output file. Format: t  y
   constant sim_out : string :=
         "simdata\resp_sqr1_30b_1g_48a_16s.txt##############";
   -- master clock
   constant clock_time : time := 0.5**26*1000 ms;   --14.9 ns;
   -- master clock high
   constant clock_hi   : time := 0.5 * clock_time;
   -- master clock low
   constant clock_low  : time := clock_time - clock_hi;
   -- ADC clock multiplier
   constant adc_clock_time : integer := 128;
   -- ADC clock multiplier high
   constant adc_clock_hi   : integer := 100;
   -- ADC clock multiplier low
   constant adc_clock_low  : integer := adc_clock_time - adc_clock_hi;
   -- 1pps clock multiplier
   constant onepps_clock_time : integer := 2**26;
   -- 1pps clock multiplier high
   constant onepps_clock_hi   : integer := 4;
   -- 1pps clock multiplier low
   constant onepps_clock_low  : integer := onepps_clock_time - onepps_clock_hi;
   -- simulation T0 time shift multiplier
   constant t0 : integer := -10;

   -- adc value type
   subtype adc_type is std_logic_vector (17 downto 0);
   -- filter output type
   subtype output_type is std_logic_vector (31 downto 0);

   -- clock signal
   signal clk : std_logic := '1';
   -- ADC signal
   signal adc_clk : std_logic := '0';
   -- 1 pps signal
   signal onepps : std_logic := '0';

   -- ADC input value
   signal x1 : adc_type := (others => '0');
   signal x1r : real := 0.0;
   -- ADC/filter overflow output value
   signal x1_ovf : std_logic := '0';
   -- latched overflow
   signal overflow : std_logic := '0';
   -- filter output value
   signal y1 : output_type := (others => '0');
   signal y1r : real := 0.0;

   -- IIR_top component
   component IIR_top is port
   (
      AD      : In    STD_LOGIC_VECTOR(11 downto 0);
      ADC0    : In    STD_LOGIC_VECTOR(17 downto 0);
      ADC0L   : In    STD_LOGIC;
      ADC0OVF : Out   STD_LOGIC;
      CLK     : In    STD_LOGIC;
      CLK1PPS : In    STD_LOGIC;
      CS      : In    STD_LOGIC;
      D       : InOut STD_LOGIC_VECTOR(31 downto 0);
      SEL     : In    STD_LOGIC_VECTOR(2 downto 0);
      WR      : In    STD_LOGIC
   );
   end component IIR_top;


begin

   ---------------------------------------------------------------------------
   -- Clock generator
   ---------------------------------------------------------------------------
   clock_gen: process is
      variable count : integer := t0;
   begin
--      clk <= '0' after clock_hi, '1' after clock_low;
      clk <= '0';
      if count mod adc_clock_time = 0 then
         adc_clk <= '1';
      elsif count mod adc_clock_time = adc_clock_hi then
         adc_clk <= '0';
      end if;
      if count mod onepps_clock_time = 0 then
         onepps <= '1';
      elsif count mod onepps_clock_time = onepps_clock_hi then
         onepps <= '0';
      end if;
      wait for clock_time/2;
      clk <= '1';
      wait for clock_time/2;
      count := count + 1;
   end process clock_gen;


   ---------------------------------------------------------------------------
   -- Stimulus: Impulse response
   ---------------------------------------------------------------------------
--   stimulus_impulse: process (clk) is
--      -- filter values
--      variable r : real := 0.0;
--   begin
--      if rising_edge (adc_clk) then
--         r := 1.0;
--         x1r <= r;
--         x1 <= convert_adc_value (r, 2.0**(adc_type'length-2));
--      end if;
--   end process stimulus_impulse;


   ---------------------------------------------------------------------------
   -- Stimulus: read excitation from file
   ---------------------------------------------------------------------------
   stimulus_file: process is
      -- filter output file
      file input : text;
      -- line
      variable l : line;
      -- status
      variable status : file_open_status;
      -- filter values
      variable r : real := 0.0;
      -- string length
      variable len : integer := sim_in'length;

      --
      -- convert a real value into a ADC value
      --
      function convert_adc_value (a : in real;
            scaling : in real := 2.0**(adc_type'length-2)) return adc_type is

         subtype adc_signed is signed (adc_type'length-1 downto 0);
         constant q : real := 2.0**(adc_type'length-1);
         constant max : adc_signed := (adc_type'left => '0', others => '1');
         variable x : real;
         variable xx : real;
         variable y : adc_signed := (others => '0');

      begin
         x := a * scaling;
         xx := abs (x);
         if (xx >= q) then
            y := max;
         else
            for i in y'left-1 downto 0 loop
               if (xx >= q/2.0) then
                  y(i) := '1';
                  xx := xx - q/2.0;
               else
                  y(i) := '0';
               end if;
               xx := 2.0 * xx;
            end loop;
            if (xx > q/2.0) and (y /= max) then
               y := y + 1;
            end if;
         end if;
         if (x < 0.0) then
            y := -y;
         end if;
         return std_logic_vector(y);
      end function convert_adc_value;

   begin
      for i in 1 to sim_in'length loop
         if sim_in (i) = '#' then
            len := i - 1;
            exit;
         end if;
      end loop;
      -- read filter coefficients from file
      file_open (status, input, path & sim_in(1 to len), read_mode);
      assert status = open_ok
         report "File open failed for " & sim_in(1 to len) severity failure;
      report "Input file is " & sim_in(1 to len);
      -- read data from disk
      while not endfile (input) loop
         readline (input, l);
         read (l, r);
         wait until rising_edge (adc_clk);
         x1r <= r;
         x1 <= convert_adc_value (r, 2.0**(adc_type'length-2));
      end loop;
      file_close (input);
      -- done
      wait;
   end process stimulus_file;


   ---------------------------------------------------------------------------
   -- Filter instantination
   ---------------------------------------------------------------------------
   filter1: component IIR_top
      port map (
         ADC0 => x1,
         ADC0L => adc_clk,
         ADC0OVF => x1_ovf,
         SEL => B"000",
         CLK => clk,
         CLK1PPS => onepps,
         AD => B"000000000001",
         D => y1,
         WR => '0',
         CS => '1'
      );

   -- convert output to real
   real_output: process (y1) is
      -- ADC output to real
      function adc_to_real (x : in std_logic_vector;
                            decimal : in integer := -1) return real is
         variable y, q : real;
         variable n : integer;
      begin
         if (decimal < 0) or (decimal > x'length) then
            n := x'length;
         else
            n := decimal;
         end if;
         q := 2.0**(n-1);
         if x(x'left) = '1' then
            y := -q;
         else
            y := 0.0;
         end if;
         for i in x'range loop
            if (i /= x'left) and (x(i) = '1') then
               y := y + q;
            end if;
            q := q / 2.0;
         end loop;
         return y;
      end function adc_to_real;
   begin
   y1r <= adc_to_real (y1, output_type'length - (adc_type'length - 2 + 9));
   end process real_output;

   -- latch overflow
   overflow <= '1' when x1_ovf = '1';


   ---------------------------------------------------------------------------
   -- Write results to disk
   ---------------------------------------------------------------------------
   recording: process is
      -- filter output file
      file output : text;
      -- line
      variable l : line;
      -- status
      variable status : file_open_status;
      -- number of data points to skip
      constant skip : integer := 16384;
      -- number of data points to keep
      constant keep : integer := 16*16384;
      -- recording enabled?
      constant enabled : boolean := true;
      -- string length
      variable len : integer := sim_out'length;

   begin
      if enabled then
         -- skip
         for i in 1 to skip loop
            wait until rising_edge (adc_clk);
         end loop;
         --
         -- read filter coefficients from file
         for i in 1 to sim_out'length loop
            if sim_out (i) = '#' then
               len := i - 1;
               exit;
            end if;
         end loop;
         file_open (status, output, path & sim_out(1 to len), write_mode);
         assert status = open_ok
            report "File open failed for " & sim_out(1 to len) severity failure;
         report "Output file is " & sim_out(1 to len);
         -- write data to disk
         for i in 1 to keep loop
            wait until rising_edge (adc_clk);
            write (l, y1r, right, 20, 14);
            writeline (output, l);
         end loop;
         file_close (output);
         report "Overflow is " & std_logic'image (overflow);
      end if;
      -- done
      wait;
   end process recording;

end architecture test;


