--
-- IIR filter types
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
-- use ieee.math_real.all;

package iir_types is
   --
   -- Biquad types and constant (real)
   --
   -- biquad coefficients
   type biquad_record is record
      a1: real;
      a2: real;
      b1: real;
      b2: real;
   end record biquad_record;
   -- maximum biquad sections
   constant biquad_max : natural := 16;
   -- array of biquads
   type biquad_array is array (natural range 1 to biquad_max) of biquad_record;
   -- iir coefficient record
   type iir_record is record
      gain: real;
      num_biquad: natural;
      biquads: biquad_array;
   end record iir_record;
   -- history array of reals
   type history_array is array (natural range 1 to biquad_max) of real;

   --
   -- Biquad types and constant (signed fixed point)
   --
   -- number of digits in a's and b's
   constant filter_coeff_digits : natural := 35;
   -- number of digits in muliplier
   constant multiplier_digits : natural := 35;
   -- number of digits in filter gain
   constant filter_gain_digits : natural := filter_coeff_digits;
   -- position of decimal point (count from left) in muliplier filter coefficient
   constant multiplier_coeff_decimal : natural := 2;
   -- position of decimal point (count from left) in muliplier filter gain
   constant multiplier_gain_decimal : natural := 2 + 0*4;
   -- number of digits in filter values
   constant filter_value_digits : natural := multiplier_digits;
   -- number of digits in final filter output
   constant filter_output_digits : natural := 32;
   -- number of digits in adc values
   constant adc_digits : natural := 18;
   -- number of digits in the accumulators
   constant accumulator_digits : natural := 73;
   -- number of dropped digits in the accumulator
   constant accumulator_dropped_digits : natural := 25;
   -- number of digits in shift value
   constant filter_shift_digits : natural := 4;
   -- right shift after converting adc to filter value
   constant adc_filter_shift : natural := 2 + (filter_value_digits - filter_output_digits);
   -- right shift after converting filter value to muliplier
   constant filter_mul_shift : natural := multiplier_digits - filter_value_digits;
   -- right shift after converting filter coefficients to muliplier
   constant coeff_mul_shift : natural := 0;
   -- right shift after converting gain to multiplier
   constant gain_mul_shift : natural := coeff_mul_shift;
   -- right shift before converting gain muliplier result back to filter value
   constant mul_filter_shift : natural := multiplier_digits-multiplier_gain_decimal;
   -- right shift after converting filter value to accumulator precision
   constant filter_accumulator_shift : natural :=
      (accumulator_digits-(multiplier_digits-multiplier_coeff_decimal)-
       (multiplier_digits-filter_mul_shift));
   -- right shift after converting muliplier result to accumulator precision
   constant mul_accumulator_shift : natural := accumulator_digits-2*multiplier_digits;
   -- right shift before converting accumulator values to filter values
   constant accumulator_filter_shift : natural :=
      accumulator_digits - filter_value_digits - filter_accumulator_shift;
   -- right shift before converting filter value to output value
   constant filter_out_shift : natural := 0;
   -- output decimation
   constant out_decimation : natural := 32;

   -- adc value type
   subtype adc_type is signed (adc_digits-1 downto 0);
   -- filter value type
   subtype filter_value_type is signed (filter_value_digits-1 downto 0);
   -- filter output type
   subtype filter_output_type is signed (filter_output_digits-1 downto 0);
   -- filter coefficient type
   subtype filter_coeff_type is signed (filter_coeff_digits-1 downto 0);
   -- multiplier input type
   subtype multiplier_type is signed (multiplier_digits-1 downto 0);
   -- multiplier output type
   subtype multiplier_result_type is signed (2*multiplier_digits-1 downto 0);
   -- accumulator type
   subtype accumulator_type is signed (accumulator_digits-accumulator_dropped_digits-1 downto 0);
   -- filter shift type
   subtype filter_shift_type is signed (filter_shift_digits-1 downto 0);
   -- biquad coefficients
   type biquad_coeffs is record
      a1: filter_coeff_type; -- sign changed to normal convention!
      a2: filter_coeff_type; -- sign changed to normal convention!
      b1: filter_coeff_type;
      b2: filter_coeff_type;
      c0: filter_shift_type;
   end record biquad_coeffs;
   -- array of biquad coeffcients
   type biquad_coeffs_array is array (natural range 1 to biquad_max) of biquad_coeffs;
  -- iir coefficient record
   type iir_coeffs is record
      gain: filter_coeff_type;
      num_biquad: natural;
      biquads: biquad_coeffs_array;
   end record iir_coeffs;
   -- biquad history
   type biquad_history is record
      x1 : filter_value_type; -- old x
      x2 : filter_value_type; -- old old x
      y1 : filter_value_type; -- old y
      y2 : filter_value_type; -- old old y
   end record biquad_history;
   -- array of biquad history values
   type iir_history is array (natural range 1 to biquad_max) of biquad_history;

   --
   --
   -- signed to hex string
   procedure signed_to_hex (str : out string; len : out integer; 
      x : in signed; digits : in integer := -1);
   -- signed to string
   -- the decimal point is specified counting from left
   procedure signed_to_string (str : out string; len : out integer; 
      x : in signed; decimal : in integer := -1);
   -- signed to real
   -- the decimal point is specified counting from left
   function signed_to_real (x : in signed; decimal : in integer := -1) return real;

   --
   -- Biquad subroutines (real)
   --
   -- computes a biquad
   procedure ideal_biquad (x : in real; y : out real;
         h1: inout real; h2: inout real; coeff: in biquad_record);

   -- computes an IIR filter on a single value
   procedure ideal_iir (x : in real; y : out real;
         h1: inout history_array; h2: inout history_array;
         iir : in iir_record);


   --
   -- Biquad subroutines (signed fixed point)
   --
   -- convert a real value into adc
   function convert_adc_value (a : in real; 
      scaling : in real := 2.0**(adc_digits-2)) return adc_type;

   -- convert a real filter coefficient into a signed
   function convert_filter_coeff (a : in real) return filter_coeff_type;

   -- convert a filter gain into a multiplier shift
   function convert_filter_gain (g : in real) return filter_shift_type;


   -- convert one signed vector to another
   -- the output array must be at least as long as the input array
   -- empty bits at the end are filled with zeros
   -- values are then shifted and sign extended if necessary
   procedure convert_vector_shift (x : in signed;
         y : out signed; shift : in natural);

   -- convert one signed vector to another
   -- the output array must be at equal or shorter than the input array
   -- the value is first shifted to the right and sign extended if necessary
   -- excess bits at the front are dropped, the overflow flag is set appropriately
   procedure convert_shift_vector (x : in signed;
         y : out signed; shift : in natural; ovf : out boolean);

   -- computes a biquad
   procedure fixedpoint_biquad (x : in filter_value_type;
         y : out filter_value_type; h: inout biquad_history;
         coeff: in biquad_coeffs; overflow : out boolean);

   -- computes an IIR filter on a single value
   procedure fixedpoint_iir (x : in adc_type;
         y : out filter_output_type; h: inout iir_history;
         iir : in iir_coeffs; overflow : out boolean;
         maxrange : out real);

end package iir_types;


package body iir_types is

   --
   -- signed to hex
   --
   procedure signed_to_hex (str : out string; len : out integer; 
      x : in signed; digits : in integer := -1) is
      variable d, n : integer;
      variable y,mask : signed (x'high downto x'low);
      variable s : string (1 to str'high-str'low+1);
      variable l : integer := 0;
   begin
      d := digits;
      y := x;
      mask := (others => '0');
      mask(x'low+0) := '1';
      mask(x'low+1) := '1';
      mask(x'low+2) := '1';
      mask(x'low+3) := '1';
      loop
         n := to_integer (y and mask);
         if (n < 10) then
            s(1 to l+1) := (character'val(character'pos('0')+n)) & s(1 to l);
         else
            s(1 to l+1) := (character'val(character'pos('A')+n-10)) & s(1 to l);
         end if;
         l := l + 1;
         y := y srl 4;
         d := d - 1;
         exit when (d = 0) or ((d < 0) and (y = 0));
      end loop;
      str := s;
      len := l;
   end procedure signed_to_hex;

   --
   -- signed to string
   --
   procedure signed_to_string (str : out string; len : out integer;
      x : in signed; decimal : in integer := -1) is
      variable y : signed (x'high downto x'low);
      variable y2 : signed (x'high+4 downto x'low) := (others => '0');
      variable s : string (1 to 100);
      variable l, l2 : integer := 0;
      variable n : integer;
   begin
      if (decimal < 0) or (decimal > x'length) then
         n := x'length;
      else
         n := decimal;
      end if;
      y := x srl (x'length - n);
      signed_to_hex (s, l, y);
      if (n /= x'length) then
         s(1 to l+1) := s(1 to l) & '.';
         l := l + 1;
         y := x;
         if n > 0 then
            y(x'high downto x'high-n+1) := (others => '0');
         end if;
         y2(x'high downto x'low) := y;
         l2 := (x'length - n) / 4;
         if (x'length - n) mod 4 /= 0 then
            y2 := y2 sll (4 - ((x'length - n) mod 4));
            l2 := l2 +1;
         end if;
         signed_to_hex (s (l+1 to 100), l2, y2, l2);
         l := l + l2;
      end if;
      str := s;
      len := l;
   end procedure signed_to_string;

   --
   -- signed to real
   --
   function signed_to_real (x : in signed; decimal : in integer := -1) return real is
      variable y, q : real;
      variable n : integer;
   begin
      if (decimal < 0) or (decimal > x'length) then
         n := x'length;
      else
         n := decimal;
      end if;
      q := 2.0**(n-1);
      if x(x'left) = '1' then
         y := -q;
      else
         y := 0.0;
      end if;
      for i in x'range loop
         if (i /= x'left) and (x(i) = '1') then
            y := y + q;
         end if;
         q := q / 2.0;
      end loop;
      return y;
   end function signed_to_real;

   --
   -- computes a biquad
   --
   procedure ideal_biquad (x : in real; y : out real;
      h1: inout real; h2: inout real; coeff: in biquad_record) is
      variable h0 : real;
   begin
      -- apply single second order section
      h0 := x - h1 * coeff.a1 - h2 * coeff.a2;
      y  := h0 + h1 * coeff.b1 + h2 * coeff.b2;
      -- update history buffers
      h2 := h1;
      h1 := h0;
   end procedure ideal_biquad;

   --
   -- computes an IIR filter on a single value
   --
   procedure ideal_iir (x : in real; y : out real;
         h1: inout history_array; h2: inout history_array;
         iir : in iir_record) is
      variable r : real;
      variable s : real;
   begin
      r := iir.gain * x;
      for i in 1 to iir.num_biquad loop
         ideal_biquad (r, s, h1(i), h2(i), iir.biquads(i));
         r := s;
      end loop;
      y := r;
   end procedure ideal_iir;


   --
   -- convert a real value into a ADC value
   --
   function convert_adc_value (a : in real; 
         scaling : in real := 2.0**(adc_digits-2)) return adc_type is
      constant q : real := 2.0**(adc_digits-1);
      constant max : adc_type := (adc_type'left => '0', others => '1');
      variable x : real;
      variable xx : real;
      variable y : adc_type := (others => '0');
   begin
      x := a * scaling;
      xx := abs (x);
      if (xx >= q) then
         y := max;
      else
         for i in y'left-1 downto 0 loop
            if (xx >= q/2.0) then
               y(i) := '1';
               xx := xx - q/2.0;
            else
               y(i) := '0';
            end if;
            xx := 2.0 * xx;
         end loop;
         if (xx > q/2.0) and (y /= max) then
            y := y + 1;
         end if;
      end if;
      if (x < 0.0) then
         y := -y;
      end if;
      return y;
   end function convert_adc_value;
   
   --
   -- convert a real filter coefficient into a signed
   --
   function convert_filter_coeff (a : in real) return filter_coeff_type is
      constant q : real := 2.0**(multiplier_coeff_decimal-2);
      constant max : filter_coeff_type := (filter_coeff_type'left => '0', others => '1');
      variable c : filter_coeff_type := (others => '0');
      variable aa : real;
      variable s : string (1 to 100);
      variable l : integer;
   begin
      aa := abs (a);
      if (aa > 2.0*q) then
         c := max;
         report "filter coefficient is too large " & real'image (a) severity error;
      else
         for i in c'left-1 downto 0 loop
            if (aa >= q) then
               c(i) := '1';
               aa := aa - q;
            else
               c(i) := '0';
            end if;
            aa := 2.0 * aa;
         end loop;
         if (aa > q) and (c /= max) then
            c := c + 1;
         end if;
      end if;
      if (a < 0.0) then
         c := -c;
      end if;
      return c;
   end function convert_filter_coeff;

   --
   -- convert a real filter gain into a multiplier shift
   --
   function convert_filter_gain (g : in real) return filter_shift_type is
      variable gain : real;
      variable shift : filter_shift_type;
      variable n : integer;
   begin
      gain := abs (g);
      n := -2**(filter_shift_digits-1);
      for i in -2**(filter_shift_digits-1) to 2**(filter_shift_digits-1)-1 loop
         if (gain > 2.0**i) then
            n := i;
         else
            exit;
         end if;
      end loop;
      shift := to_signed (n, filter_shift_digits);
      return shift;
   end function convert_filter_gain;

   --
   -- convert one std vector to another
   --
   procedure convert_vector_shift (x : in signed;
         y : out signed; shift : in natural) is
      variable temp : signed (y'left downto y'right);
   begin
      if y'left - y'right >= x'left - x'right then
         temp(y'left downto (y'left-(x'left-x'right))) := x;
         temp((y'left-x'left-1) downto y'right) := (others => '0');
      else
         temp(y'left downto y'right) := x(x'left downto (x'left-(y'left-y'right)));
      end if;
      y := shift_right (temp, shift);
   end procedure convert_vector_shift;

   --
   -- convert one std vector to another
   --
   procedure convert_shift_vector (x : in signed;
         y : out signed; shift : in natural; ovf : out boolean) is
      variable temp : signed (x'left downto x'right);
      variable b : boolean := false;
   begin
      temp := shift_right (x, shift);
      y := temp(y'left+x'right downto x'right);
      for i in x'left downto y'left+x'right+1 loop
         b := b or (temp(i) /= temp(y'left));
      end loop;
      ovf := b;
   end procedure convert_shift_vector;

   --
   -- computes a biquad
   --
   procedure fixedpoint_biquad (x : in filter_value_type;
         y : out filter_value_type; h: inout biquad_history;
         coeff: in biquad_coeffs; overflow : out boolean) is
      variable y0 : filter_value_type;
      variable x1, x2, y1, y2 : multiplier_type;
      variable a1, a2, b1, b2 : multiplier_type;
      variable bx1, bx2, ay1, ay2 : multiplier_result_type;
      variable s, s0, s1, s2, s3, s4 : accumulator_type;
      variable ovf : boolean;
   begin
      -- convert filter values and filter coefficients to muliplier type
      convert_vector_shift (h.x1, x1, filter_mul_shift);
      convert_vector_shift (h.x2, x2, filter_mul_shift);
      convert_vector_shift (h.y1, y1, filter_mul_shift);
      convert_vector_shift (h.y2, y2, filter_mul_shift);
      a1 := coeff.a1;
      a2 := coeff.a2;
      b1 := coeff.b1;
      b2 := coeff.b2;
      -- do multiplication
      bx1 := b1 * x1;
      bx2 := b2 * x2;
      ay1 := a1 * y1;
      ay2 := a2 * y2;
      -- convert mulplier result to accumulator precision
      convert_vector_shift (x, s0, filter_accumulator_shift);
      convert_vector_shift (bx1, s1, mul_accumulator_shift);
      convert_vector_shift (bx2, s2, mul_accumulator_shift);
      convert_vector_shift (ay1, s3, mul_accumulator_shift);
      convert_vector_shift (ay2, s4, mul_accumulator_shift);
      -- sum over input values
      s := s0 + s1 + s2;
      -- scaling
      if (coeff.c0 > 0) then
         s := shift_left (s, to_integer (coeff.c0));
      else
         s := shift_right (s, -to_integer(coeff.c0));
      end if;
      -- add ouput values
      s := s + s3 + s4;
      -- convert sum back to filter value
      convert_shift_vector (s, y0, accumulator_filter_shift-accumulator_dropped_digits, ovf);
      y := y0;
      overflow := ovf;
      -- update history
      h.x2 := h.x1;
      h.x1 := x;
      h.y2 := h.y1;
      h.y1 := y0;
   end procedure fixedpoint_biquad;

   --
   -- computes an IIR filter on a single value
   --
   procedure fixedpoint_iir (x : in adc_type;
         y : out filter_output_type; h: inout iir_history;
         iir : in iir_coeffs; overflow : out boolean;
         maxrange : out real) is
      constant maximum : adc_type := ('0', others => '1');
      constant minimum : adc_type := ('1', others => '0');
      
      function calc_range (x : in signed) return real is
         variable xr : real;
      begin
         xr := abs (signed_to_real (x, 1));
         return xr;
      end function calc_range;
      
      variable x0, y0 : filter_value_type;
      variable xm, gain : multiplier_type;
      variable gx0 : multiplier_result_type;
      variable ovf, o : boolean := false;
      variable rang, mrang : real;
   begin
      -- overflow
      ovf := (x = maximum) or (x = minimum);
      mrang := 0.0;
      -- gain multiplication
      convert_vector_shift (x, x0, adc_filter_shift);
      rang := calc_range (x0);
      if rang > mrang then
         mrang := rang;
      end if;
      convert_vector_shift (x0, xm, filter_mul_shift);
      convert_vector_shift (iir.gain, gain, gain_mul_shift);
      gx0 := xm * gain;
      convert_shift_vector (gx0, x0, mul_filter_shift, o);
      ovf := ovf or o;
      rang := calc_range (x0);
      if rang > mrang then
         mrang := rang;
      end if;
      -- iterate over second order sections
      for i in 1 to iir.num_biquad loop
         fixedpoint_biquad (x0, y0, h(i), iir.biquads(i), o);
         x0 := y0;
         ovf := ovf or o;
         rang := calc_range (x0);
         if rang > mrang then
            mrang := rang;
         end if;
      end loop;
      convert_shift_vector (x0, y, filter_out_shift, o);
      overflow := ovf or o;
      maxrange := mrang;
   end procedure fixedpoint_iir;

end package body iir_types;

